@@ -3,7 +3,16 @@ import importlib
 import importlib.util
 import inspect
 import itertools
-from typing import Any, Callable, Optional, Type, TypeVar, Union
+from typing import (
+    Any,
+    Callable,
+    Generic,
+    Optional,
+    Type,
+    TypeVar,
+    Union,
+    overload,
+)
 
 from omegaconf import DictConfig
 
@@ -31,12 +40,12 @@ class InitLater(functools.partial):
         cl.type_ = _type_
         return cl
 
-    def get_kw(self: "InitLater", *args, **kwargs):
+    def get_kw(self: "InitLater", *args, **kwargs) -> Any:
         """Shortcut for accessing parameters that have been
         provided to an InitLater object"""
         return self.keywords.get(*args, **kwargs)
 
-    def pop_kw(self: "InitLater", *args, **kwargs):
+    def pop_kw(self: "InitLater", *args, **kwargs) -> Any:
         """Shortcut for popping parameters that have been
         provided to an InitLater object"""
         return self.keywords.pop(*args, **kwargs)
@@ -186,16 +195,17 @@ class Target:
 
 
 InitT = TypeVar("InitT")
+InitNestedT = TypeVar("InitNestedT")
 CallableT = TypeVar("CallableT", bound=Callable)
 
 
-class init:
+class init(Generic[InitT, CallableT]):
     TARGET: str = "_target_"
 
     @classmethod
     def callable(
         cls: Type["init"],
-        config: Optional[Any] = None,
+        _config_: Optional[Any] = None,
         _type_: Optional[CallableT] = None,
         _recursive_: Optional[Any] = None,
         target: Optional[str] = None,
@@ -221,21 +231,47 @@ class init:
             A callable function that can be used to instantiate an object.
         """
 
-        if target is not None and config is not None:
+        if target is not None and _config_ is not None:
             raise ValueError("Cannot specify both `target` and `config`")
         elif target is None:
-            config = cast(config)
+            _config_ = cast(_config_)
             try:
-                target = str(config[cls.TARGET])
+                target = str(_config_[cls.TARGET])
             except KeyError:
-                raise KeyError(f"Config `{config}` has no `{cls.TARGET}` key!")
+                raise KeyError(
+                    f"Config `{_config_}` has no `{cls.TARGET}` key!"
+                )
 
         return Target.from_string(target)
 
+    @overload
+    @classmethod
+    def later(
+        cls: Type["init"],
+        _config_: Any,
+        _type_: Type[InitT],
+        _recursive_: bool = True,
+        /,
+        **kwargs: Any,
+    ) -> Callable[..., InitT]:
+        ...
+
+    @overload
+    @classmethod
+    def later(
+        cls: Type["init"],
+        _config_: Optional[Any] = None,
+        _type_: Optional[Type[InitT]] = None,
+        _recursive_: bool = True,
+        /,
+        **kwargs: Any,
+    ) -> Any:
+        ...
+
     @classmethod
     def later(
         cls: Type["init"],
-        config: Optional[Any] = None,
+        _config_: Optional[Any] = None,
         _type_: Optional[Type[InitT]] = None,
         _recursive_: bool = True,
         /,
@@ -265,16 +301,16 @@ class init:
         SpringsWarnings.missing_type(fn_name="init.later", type_=_type_)
 
         # if no config is provided, we return a function
-        if config is None:
+        if _config_ is None:
             return InitLater.no_op()
 
         # ensure we are working with a ConfigDict
-        config_node = cast(config)
+        config_node = cast(_config_)
 
         # try to get the target callable from either the config
         # or keyword arguments passed to the init function
         if cls.TARGET in config_node:
-            fn = cls.callable(config=config_node, _type_=_type_)
+            fn = cls.callable(_config_=config_node, _type_=_type_)
         elif cls.TARGET in kwargs:
             fn = cls.callable(target=kwargs[cls.TARGET], _type_=_type_)
         else:
@@ -283,28 +319,118 @@ class init:
                 f"`{kwargs}`: `{cls.TARGET}` keyword missing"
             )
 
-        def _recursive_init(param):
+        def _recursive_init(
+            param: Union[dict, DictConfig, None], type_: Optional[type] = None
+        ) -> Any:
+            """Given one of the parameters for the object that we are
+            trying to instantiate, plus optionally its expected type, this
+            function will try to instantiate the parameter if it is a nested
+            config AND if _recursive_ is True.
+            """
             must_recursive_init = (
+                # we are operating in recursive mode
                 _recursive_
+                # the parameter is a nested config
                 and isinstance(param, (DictConfig, dict))
+                # the nested config has a TARGET property
                 and cls.TARGET in param
             )
             if must_recursive_init:
-                param = cls.later(config=param, _recursive_=True)
-            return param
+                # we initialize recursively
+                return cls.later(param, type_, _recursive_)
+            else:
+                # we simply return the parameter
+                return param
+
+        def _find_child_type(
+            cls_: Union[None, type], attr_name: str
+        ) -> Union[None, type]:
+            """Given the type the object we are initializing is expected to be,
+            as well as the name of the attribute we are about to pass to its
+            initializer, this function figures out the expected type using
+            a combination of class annotations and __init__ annotations.
+
+            If the type the attribute should be cannot be determined, it
+            simply returns None.
+
+            An example:
+
+            ```python
+            class Test1:
+                def __init__(self, a: int):
+                    ...
+
+            print(_find_child_type(Test1, "a")) # prints "int"
+
+            class Test2:
+                a: float
+
+                def __init__(self, a):
+                    ...
+
+            print(_find_child_type(Test2, "a")) # prints "float"
+
+            class Test3:
+                def __init__(self, a):
+                    ...
+
+            print(_find_child_type(Test3, "a")) # prints "None"
+            ```
+            """
+
+            if cls_ is None:
+                # if the class is None, we don't perform any checks
+                return None
+
+            if attr_name in (parent_anns := inspect.get_annotations(cls_)):
+                return parent_anns[attr_name]
+
+            if attr_name in (
+                parent_anns := inspect.getfullargspec(cls_).annotations
+            ):
+                return parent_anns[attr_name]
+
+            # this is the case where we cannot resolve anything
+            return None
 
         init_call_dict = {
-            str(k): _recursive_init(v)
+            str(k): _recursive_init(
+                param=v, type_=_find_child_type(cls_=_type_, attr_name=str(k))
+            )
             for k, v in {**config_node, **kwargs}.items()
             if k != cls.TARGET
         }
 
         return InitLater(fn, _type_, **init_call_dict)
 
+    @overload
     @classmethod
     def now(
         cls: Type["init"],
-        config: Optional[Any] = None,
+        _config_: Any,
+        _type_: Type[InitT],
+        _recursive_: bool = True,
+        /,
+        **kwargs: Any,
+    ) -> InitT:
+        ...
+
+    @overload
+    @classmethod
+    def now(
+        cls: Type["init"],
+        _config_: Optional[Any] = None,
+        _type_: Optional[Type[InitT]] = None,
+        _recursive_: bool = True,
+        /,
+        **kwargs: Any,
+    ) -> Any:
+        ...
+
+    @classmethod
+    def now(
+        cls: Type["init"],
+        _config_: Optional[Any] = None,
         _type_: Optional[Type[InitT]] = None,
         _recursive_: bool = True,
         /,
@@ -335,22 +461,9 @@ class init:
         # in its signature, which is required to ensure that **kwargs
         # works well with type annotations.
         init_call: Callable[..., InitT] = cls.later(
-            config, _type_, _recursive_, **kwargs
+            _config_, _type_, _recursive_, **kwargs
         )
         return init_call()
 
-    def __new__(  # type: ignore
-        cls: Type["init"],
-        config: Optional[Any] = None,
-        _type_: Optional[Type[InitT]] = None,
-        _recursive_: bool = True,
-        /,
-        **kwargs: Any,
-    ) -> InitT:
-        """Convenience shortcut for `init.now`."""
-
-        # notice the use of non-keyword arguments here for config,
-        # _type_, and _recursive_. This is because `now` has a `/`
-        # in its signature, which is required to ensure that **kwargs
-        # works well with type annotations.
-        return cls.now(config, _type_, _recursive_, **kwargs)
+    """Convenience shortcut for `init.now`"""
+    __new__: Callable[..., InitT] = now  # type: ignore
